/*
debug_malloc.c 1.1

 Copyright 1993-1997 Jaakko Hyvtti
Email: Jaakko.Hyvatti@iki.fi
Phone: +358 40 5011222
URL:   http://www.iki.fi/hyvatti/

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

The GNU C Library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; see the file COPYING.LIB.  If
not, write to the author.
*/

/*
1993

If you are looking for tools to examine memory leaks, others surely
will point you to more complete packages, but let me hereby announce a
small tool I personally use.  It just logs all malloc()s and free()s
and keeps a database to trap errors.  After your program is finished,
you may examine the produced logfile to find out unfreed malloc()s.
It's a bit of a hackers toy, you need to do some work with gdb and
breakpoints to find where the actual leak is.

Just link it with your program using -ggdb, and initialize debugging by 
calling init_debug_malloc("debugoutputfilename", 1); analyze later by
looking at file and with accompanying debug_malloc.awk:

awk -f debug_malloc.awk debugoutputfilename

Someone should write code to do some stack trace on every malloc and free.
Anyway, I have successfully debugged finger, wu-ftpd, Linux-libc etc.
for leaks and bad calls to free() with this as it is.

1995-08-17

Made available at ftp://www.fi

1995-11-01

New with version 1.1: a call backtrace from stack is dumped.

1997-02-16

Made available at http://www.iki.fi/hyvatti/sw/debug_malloc-1.1.tar.gz
*/

#include <stdio.h>
#include <malloc.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>

#ifdef JHLIB
#include "jh.h"
#endif

#ifndef DEBUGFILE
#define DEBUGFILE "malloc.debug"
#endif

#ifndef STOP
#define STOP sleep (1) /* We'll pause in a bad free() */
#endif

#ifndef HASH
#define HASH 200003 /* a prime number */
#endif

static void free_hook (__ptr_t osoitin);
static __ptr_t malloc_hook (size_t koko);
static __ptr_t realloc_hook (__ptr_t osoitin, size_t koko);

static FILE *debugfile = NULL;
static int print_stack = 0;

struct db {
  __ptr_t addr;
  size_t size;
  struct db *next;
};

static struct db db [HASH];

void
init_debug_malloc (const char *nimi, int ps)
{
  if (NULL == debugfile) {
    if (nimi)
      if (!strcmp ("-", nimi))
	debugfile = stderr;
      else
	debugfile = fopen (nimi, "w");
    if (NULL == debugfile)
      debugfile = fopen (DEBUGFILE, "w");
  }
  if (NULL != debugfile) {
    __free_hook = free_hook;
    __malloc_hook = malloc_hook;
    __realloc_hook = realloc_hook;
    memset (db, 0, sizeof (db));
  }
  print_stack = ps;
}

/*
   osoitteet nyttvt olevan aina 8:n kerrannaisia
*/
static int
hash (__ptr_t a)
{
  return ((int)a/8 + (int)a/3) % HASH;
}

static void
add_db (__ptr_t a, size_t s)
{
  int h = hash (a);
  struct db *new;

  if (db [h].addr) {
    new = malloc (sizeof (struct db));
    new->next = db [h].next;
    db [h].next = new;
  }
  else
    new = &db [h];
  new->addr = a;
  new->size = s;
}


static void
remove_db (__ptr_t a)
{
  int h = hash (a);
  struct db *tmp, *tmp2;

  if (a == db [h].addr) {
    if (db [h].next) {
      tmp = db [h].next;
      memcpy (&db [h], tmp, sizeof (struct db));
      free (tmp);
    }
    else
      db [h].addr = NULL;
  }
  else {
    tmp = &db [h];
    while (a != tmp->next->addr)
      tmp = tmp->next;
    tmp2 = tmp->next;
    tmp->next = tmp->next->next;
    free (tmp2);
  }
}


static struct db *
lookup_db (__ptr_t a)
{
  struct db *tmp = &db [hash (a)];

  while (tmp && tmp->addr != a)
    tmp = tmp->next;
  return tmp;
}

static void
dprint (char *fmt, ...)
{
  unsigned long *st;
  va_list args;

  va_start (args, fmt);
  vfprintf (debugfile, fmt, args);
  va_end (args);
  st = (unsigned long *)&st;
  fprintf (debugfile, " [%08lX]", (unsigned long)st);
  if (print_stack)
    while ((unsigned long)++st & 4095) {

      /* Now this presumes the text segment is 2 MB.  This is of course
	 wrong but will do for now.  Some data pointers will get printed
	 if the text segment is smaller and some call trace will not get
	 printed if the segment is bigger.  Of course this also presumes
	 linux ELF binary format etc.  */

      if (*st >= 0x08000000 && *st < 0x08200000)
	fprintf (debugfile, " %08lX", *st);
    }
  fputc ('\n', debugfile);
}

static void
free_hook (__ptr_t osoitin)
{
  struct db *p = NULL;

  __free_hook = NULL;
  __malloc_hook = NULL;
  __realloc_hook = NULL;
  if (NULL == osoitin || NULL == (p = lookup_db (osoitin))) {
    STOP;
  }
  else
    remove_db (osoitin);
  free (osoitin);
  if (p)
    dprint ("free %08x %d", (unsigned int)osoitin, p->size);
  else
    dprint ("free %08x *ERROR*", (unsigned int)osoitin);
  fflush (debugfile);
  __free_hook = free_hook;
  __malloc_hook = malloc_hook;
  __realloc_hook = realloc_hook;
}

static __ptr_t
malloc_hook (size_t koko)
{
  __ptr_t tulos;
  struct db *p = NULL;

  __free_hook = NULL;
  __malloc_hook = NULL;
  __realloc_hook = NULL;
  tulos = malloc (koko);
  if (NULL == tulos || NULL != (p = lookup_db (tulos))) {
    STOP;
  }
  else
    add_db (tulos, koko);
  if (p) {
    dprint ("malloc %d == %08x *ERROR* %d", koko, (unsigned int)tulos, p->size);
    p->size = koko;
  }
  else if (tulos)
    dprint ("malloc %d == %08x", koko, (unsigned int)tulos);
  else
    dprint ("malloc %d == %08x *ERROR*", koko, (unsigned int)tulos);
  fflush (debugfile);
  __free_hook = free_hook;
  __malloc_hook = malloc_hook;
  __realloc_hook = realloc_hook;
  return tulos;
}

static __ptr_t
realloc_hook (__ptr_t osoitin, size_t koko)
{
  __ptr_t tulos;
  struct db *p = NULL, *q = NULL;

  __free_hook = NULL;
  __malloc_hook = NULL;
  __realloc_hook = NULL;
  if (NULL == osoitin || NULL == (p = lookup_db (osoitin))) {
    STOP;
  }
  tulos = realloc (osoitin, koko);
  if (NULL == tulos || (osoitin != tulos && NULL != (q = lookup_db (tulos)))) {
    STOP;
  }
  if (p) {
    if (q) {
      dprint ("realloc %08x %d == %08x *ERROR* %d", (unsigned int)osoitin, koko, (unsigned int)tulos, q->size);
      remove_db (osoitin);
      q->size = koko;
    }
    else if (tulos) {
      dprint ("realloc %08x %d == %08x", (unsigned int)osoitin, koko, (unsigned int)tulos);
      if (osoitin == tulos)
	p->size = koko;
      else {
	remove_db (osoitin);
	add_db (tulos, koko);
      }
    }
    else {
      dprint ("realloc %08x %d == %08x *ERROR*", (unsigned int)osoitin, koko, (unsigned int)tulos);
      remove_db (osoitin);
    }
  }
  else {
    if (q) {
      dprint ("realloc %08x *ERROR* %d == %08x *ERROR* %d", (unsigned int)osoitin, koko, (unsigned int)tulos, q->size);
      q->size = koko;
    }
    else if (tulos) {
      dprint ("realloc %08x *ERROR* %d == %08x", (unsigned int)osoitin, koko, (unsigned int)tulos);
      add_db (tulos, koko);
    }
    else {
      dprint ("realloc %08x *ERROR* %d == %08x *ERROR*", (unsigned int)osoitin, koko, (unsigned int)tulos);
    }
  }
    
  fflush (debugfile);
  __free_hook = free_hook;
  __malloc_hook = malloc_hook;
  __realloc_hook = realloc_hook;
  return tulos;
}

